import json
import logging
import os

import numpy as np
from keras.callbacks import ModelCheckpoint, CSVLogger, EarlyStopping
from keras.models import load_model

import arguments
import dataset
import hyperband
import models

# create logger
logger = logging.getLogger('logger')
logger.setLevel(logging.DEBUG)

# create console handler and set level to debug
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)

# create formatter
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

# add formatter to ch
ch.setFormatter(formatter)

# add ch to logger
logger.addHandler(ch)

ch = logging.FileHandler("info.log")
ch.setLevel(logging.DEBUG)

# create formatter
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

# add formatter to ch
ch.setFormatter(formatter)

# add ch to logger
logger.addHandler(ch)

args = arguments.parse()


# config = tf.ConfigProto(log_device_placement=True)
# config.gpu_options.per_process_gpu_memory_fraction = 0.5
# config.gpu_options.allow_growth = True
# session = tf.Session(config=config)


def exists_or_make_path(p):
    if not os.path.exists(p):
        os.makedirs(p)


def main_paul_best():
    char_dict = dataset.get_character_dict()
    domain_tr, flow_tr, client_tr, server_tr = load_or_generate_h5data(args.h5data, args.train_data,
                                                                       args.domain_length, args.window)

    param = models.pauls_networks.best_config
    param["vocab_size"] = len(char_dict) + 1

    embedding, model = models.get_models_by_params(param)

    model.compile(optimizer='adam',
                  loss='categorical_crossentropy',
                  metrics=['accuracy'])

    model.fit([domain_tr, flow_tr],
              [client_tr, server_tr],
              batch_size=args.batch_size,
              epochs=args.epochs,
              shuffle=True,
              validation_split=0.2)

    embedding.save(args.embedding_model)
    model.save(args.clf_model)


def main_hyperband():
    char_dict = dataset.get_character_dict()
    user_flow_df = dataset.get_user_flow_data(args.train_data)

    params = {
        # static params
        "type": ["paul"],
        "batch_size": [args.batch_size],
        "vocab_size": [len(char_dict) + 1],
        "window_size": [10],
        "domain_length": [40],
        "flow_features": [3],
        "input_length": [40],
        # model params
        "embedding_size": [16, 32, 64, 128, 256, 512],
        "filter_embedding": [16, 32, 64, 128, 256, 512],
        "kernel_embedding": [1, 3, 5, 7, 9],
        "hidden_embedding": [16, 32, 64, 128, 256, 512],
        "dropout": [0.5],
        "domain_features": [16, 32, 64, 128, 256, 512],
        "filter_main": [16, 32, 64, 128, 256, 512],
        "kernels_main": [1, 3, 5, 7, 9],
        "dense_main": [16, 32, 64, 128, 256, 512],
    }
    param = hyperband.sample_params(params)
    logger.info(param)

    logger.info("create training dataset")
    domain_tr, flow_tr, client_tr, server_tr = dataset.create_dataset_from_flows(user_flow_df, char_dict,
                                                                                 max_len=args.domain_length,
                                                                                 window_size=args.window)

    hp = hyperband.Hyperband(params,
                             [domain_tr, flow_tr],
                             [client_tr, server_tr])
    results = hp.run()
    json.dump(results, open("hyperband.json"))


def load_or_generate_h5data(h5data, train_data, domain_length, window_size):
    char_dict = dataset.get_character_dict()
    logger.info(f"check for h5data {h5data}")
    try:
        open(h5data, "r")
    except FileNotFoundError:
        logger.info("h5 data not found - load csv file")
        user_flow_df = dataset.get_user_flow_data(train_data)
        logger.info("create training dataset")
        domain_tr, flow_tr, client_tr, server_tr = dataset.create_dataset_from_flows(user_flow_df, char_dict,
                                                                                     max_len=domain_length,
                                                                                     window_size=window_size)
        logger.info("store training dataset as h5 file")
        dataset.store_h5dataset(args.h5data, domain_tr, flow_tr, client_tr, server_tr)
    logger.info("load h5 dataset")
    return dataset.load_h5dataset(h5data)


def main_train():
    exists_or_make_path(args.model_path)

    char_dict = dataset.get_character_dict()
    domain_tr, flow_tr, client_tr, server_tr = load_or_generate_h5data(args.h5data, args.train_data,
                                                                       args.domain_length, args.window)

    # parameter
    param = {
        "type": "paul",
        "batch_size": 64,
        "window_size": args.window,
        "domain_length": args.domain_length,
        "flow_features": 3,
        "vocab_size": len(char_dict) + 1,
        #
        'dropout': 0.5,
        'domain_features': args.domain_embedding,
        'embedding_size': args.embedding,
        'filter_main': 128,
        'flow_features': 3,
        'dense_main': 512,
        'filter_embedding': args.hidden_char_dims,
        'hidden_embedding': args.domain_embedding,
        'kernel_embedding': 3,
        'kernels_main': 3,
        'input_length': 40
    }

    embedding, model = models.get_models_by_params(param)
    embedding.summary()
    model.summary()
    logger.info("define callbacks")
    cp = ModelCheckpoint(filepath=args.clf_model,
                         monitor='val_loss',
                         verbose=False,
                         save_best_only=True)
    csv = CSVLogger(args.train_log)
    early = EarlyStopping(monitor='val_loss',
                          patience=5,
                          verbose=False)
    logger.info("compile model")
    model.compile(optimizer='adam',
                  loss='categorical_crossentropy',
                  metrics=['accuracy'])
    logger.info("start training")
    model.fit([domain_tr, flow_tr],
              [client_tr, server_tr],
              batch_size=args.batch_size,
              epochs=args.epochs,
              callbacks=[cp, csv, early],
              shuffle=True,
              validation_split=0.2)
    logger.info("save embedding")
    embedding.save(args.embedding_model)


def main_test():
    domain_val, flow_val, client_val, server_val = load_or_generate_h5data(args.h5data, args.train_data,
                                                                           args.domain_length, args.window)
    clf = load_model(args.clf_model)
    loss, _, _, client_acc, server_acc = clf.evaluate([domain_val, flow_val],
                                                      [client_val, server_val],
                                                      batch_size=args.batch_size)
    logger.info(f"loss: {loss}\nclient acc: {client_acc}\nserver acc: {server_acc}")
    y_pred = clf.predict([domain_val, flow_val],
                         batch_size=args.batch_size)
    np.save(os.path.join(args.model_path, "future_predict.npy"), y_pred)


def main_visualization():
    mask = dataset.load_mask_eval(args.data, args.test_image)
    y_pred_path = args.model_path + "pred.npy"
    logger.info("plot model")
    model = load_model(args.model_path + "model.h5",
                       custom_objects=evaluation.get_metrics())
    visualize.plot_model(model, args.model_path + "model.png")
    logger.info("plot training curve")
    logs = pd.read_csv(args.model_path + "train.log")
    visualize.plot_training_curve(logs, "{}/train.png".format(args.model_path))
    pred = np.load(y_pred_path)
    logger.info("plot pr curve")
    visualize.plot_precision_recall(mask, pred, "{}/prc.png".format(args.model_path))
    visualize.plot_precision_recall_curves(mask, pred, "{}/prc2.png".format(args.model_path))
    logger.info("plot roc curve")
    visualize.plot_roc_curve(mask, pred, "{}/roc.png".format(args.model_path))
    logger.info("store prediction image")
    visualize.save_image_as(pred, "{}/pred.png".format(args.model_path))


def main_score():
    mask = dataset.load_mask_eval(args.data, args.test_image)
    pred = np.load(args.pred)
    visualize.score_model(mask, pred)


def main():
    if "train" in args.modes:
        main_train()
    if "hyperband" in args.modes:
        main_hyperband()
    if "test" in args.modes:
        main_test()
    if "fancy" in args.modes:
        main_visualization()
    if "score" in args.modes:
        main_score()
    if "paul" in args.modes:
        main_paul_best()


if __name__ == "__main__":
    main()
